package br.unirio.processo.general.metaheuristics;

import java.util.Iterator;
import java.util.Vector;

import jmetal.util.PseudoRandom;
import br.unirio.processo.general.model.component.ComponenteProcesso;
import br.unirio.processo.general.model.component.RepositorioComponentes;
import br.unirio.processo.general.model.process.Artefato;
import br.unirio.processo.general.model.process.Processo;

public class ProcessCalculator
{
	private static final Integer[] pesosPraticas = { 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4 };

	public RepositorioComponentes repositorio;
	public Processo processo;
	private Vector<ComponenteProcesso> cpmandatorios;
	private Vector<ComponenteProcesso> cpinviaveis;
	private Vector<ComponenteProcesso> cpotimizacao;

	/**
	 * Constructor
	 */
	public ProcessCalculator(RepositorioComponentes repositorio, Processo processo)
	{
		this.repositorio = repositorio;
		this.processo = processo;
		this.cpmandatorios = repositorio.montaListaComponentesMandatorios();
		this.cpinviaveis = repositorio.montaListaComponentesInviaveis(processo);
		this.cpotimizacao = repositorio.montaListaComponentesOtimizacao(cpmandatorios, cpinviaveis);
	}

	/**
	 * Auxiliar - Calcula o total de artefatos produzidos pela soluo completa
	 */
	private double calculaTotalArtefatosProduzidosSolucao(int[] cromossomoComponentes)
	{
		double artefatosProduzidosSolucao = 0;
		Vector<Artefato> artefatos = new Vector<Artefato>();

		for (int i = 0; i < cromossomoComponentes.length; i++)
		{
			ComponenteProcesso cp = repositorio.getComponenteIndex(i);
			Iterator<Artefato> artefatosSaida;
			artefatosSaida = cp.getArtefatosSaida();
		
			while (artefatosSaida.hasNext())
			{
				Artefato artef = artefatosSaida.next();
			
				if (processo.checaProdutosEsperados(artef) && !artefatos.contains(artef))
					artefatos.add(artef);
			}
		}
	
		artefatosProduzidosSolucao = artefatos.size();
		return artefatosProduzidosSolucao;
	}

	/**
	 * Mtrica M1: percentual dos artefatos finais gerados
	 */
	private double calculaPercentualArtefatosFinaisGerados(int[] cromossomoComponentes)
	{
		double artefatosProduzidosProcesso = processo.getProdutoEsperadoCount();

		if (artefatosProduzidosProcesso == 0)
			return 0;

		double artefatosProduzidosSolucao = calculaTotalArtefatosProduzidosSolucao(cromossomoComponentes);
		return artefatosProduzidosSolucao / artefatosProduzidosProcesso;
	}

	/**
	 * Colaborao do Processo todo
	 */
	private double calculaPotencialColaboracao(int[] cromossomoComponentes)
	{
		int somatorio = 0;

		for (int i = 0; i < cromossomoComponentes.length; i++)
		{
			if (cromossomoComponentes[i] == 1)
			{
				ComponenteProcesso cp = repositorio.getComponenteIndex(i);
				Boolean[] praticasColaboracaoComponente = cp.getPraticasColaboracao();

				for (int j = 0; j < praticasColaboracaoComponente.length; j++)
					if (praticasColaboracaoComponente[j])
						somatorio += pesosPraticas[j];
			}
		}

		return somatorio;
	}

	/**
	 * Converte a solucao para contemplar todos os componentes
	 */
	private int[] converteSolucaoArray(int[] solucao)
	{
		int[] componentes = new int[cpotimizacao.size() + cpmandatorios.size() + cpinviaveis.size()];
		
		for (ComponenteProcesso cp : cpmandatorios)
		{
			int indice = repositorio.getIndiceComponente(cp);
			componentes[indice] = 1;
		}

		for (int i = 0; i < cpotimizacao.size(); i++)
		{
			ComponenteProcesso cp = cpotimizacao.get(i);
			int indice = repositorio.getIndiceComponente(cp);
			componentes[indice] = solucao[i];
		}
		
		return componentes;
	}
	
	/**
	 * Calcula o Fitness da soluo
	 */
	public double calculaFitness(int[] solucao)
	{
		int[] solucaoCompleta = converteSolucaoArray(solucao);
		double m1 = calculaPercentualArtefatosFinaisGerados(solucaoCompleta);
		return (m1 == 1) ? calculaPotencialColaboracao(solucaoCompleta) : 0.0;
	}

	/**
	 * Imprime total de solues no console para acompanhamento durante a execuo dos algoritmos
	 */
	public String imprimeSolucoes(int[] solucao, int ciclo, long executionTime, double fitness)
	{
		StringBuffer saida = new StringBuffer("");
		saida.append("Ciclo: " + ciclo);
		saida.append("; ExecTime: " + executionTime);
		saida.append("; Fitness: " + fitness);
		saida.append("; Soluo: {");

		for (int j = 0; j < solucao.length; j++)
			saida.append(solucao[j]);

		saida.append("}");
		return saida.toString();
	}

	/**
	 * Generates random solution
	 */
	public int[] geraSolucaoAleatoria()
	{
		int tamanho = cpotimizacao.size();
		int vetorAletatorio[] = new int[tamanho];

		for (int i = 0; i < tamanho; i++)
			vetorAletatorio[i] = PseudoRandom.randInt(0, 1);

		return vetorAletatorio;
	}

	/**
	 * Copies a solution from one array to another
	 */
	public void copySolution(int[] currentSolution, int[] bestSolution)
	{
		for (int i = 0; i < currentSolution.length; i++)
			bestSolution[i] = currentSolution[i];
	}

	/**
	 * Returns the size of a solution
	 */
	public int getSolutionSize()
	{
		return cpotimizacao.size();
	}
}