package br.unirio.processo.general.metaheuristics;

import br.unirio.processo.general.model.component.RepositorioComponentes;
import br.unirio.processo.general.model.process.Processo;

/**
 * Hill Climbing searcher for the process problem
 */
public class ProcessHillClimbing
{
	/**
	 * Calculator instance
	 */
	private ProcessCalculator calculator;

	/**
	 * Current number of evaluations
	 */
	private int evaluations;

	/**
	 * Initializes the Hill Climbing search process
	 */
	public ProcessHillClimbing(RepositorioComponentes repositorio, Processo processo)
	{
		this.calculator = new ProcessCalculator(repositorio, processo);
	}

	/**
	 * Creates a clone from a solution
	 */
	private int[] cloneSolution(int[] currentSolution)
	{
		int[] newSolution = new int[currentSolution.length];

		for (int i = 0; i < currentSolution.length; i++)
			newSolution[i] = currentSolution[i];

		return newSolution;
	}

	/**
	 * Runs a neighborhood visit starting from a given solution
	 */
	private NeighborhoodVisitorResult visitNeighbors(int[] baseSolution, int maxEvaluations)
	{
		int[] bestNeighborSolution = null;
		double bestNeighborFitness = calculator.calculaFitness(baseSolution);
		int[] neighborSolution = cloneSolution(baseSolution);

		for (int i = 0; i < baseSolution.length; i++)
		{
			neighborSolution[i] = (baseSolution[i] == 0) ? 1 : 0;
			double neighborFitness = calculator.calculaFitness(neighborSolution);

			if (neighborFitness > bestNeighborFitness)
			{
				bestNeighborFitness = neighborFitness;
				bestNeighborSolution = cloneSolution(neighborSolution);
			}

			if (++evaluations >= maxEvaluations)
				break;
		}

		if (bestNeighborSolution == null)
			return new NeighborhoodVisitorResult(NeighborhoodVisitorStatus.NO_BETTER_NEIGHBOR);

		return new NeighborhoodVisitorResult(NeighborhoodVisitorStatus.FOUND_BETTER_NEIGHBOR, bestNeighborFitness, bestNeighborSolution);
	}

	/**
	 * Gera uma soluo inicial Aleatria Vlida
	 */
	private int[] geraSolucaoAleatoriaValida()
	{
		int[] solucao = calculator.geraSolucaoAleatoria();

		while (calculator.calculaFitness(solucao) == 0.0)
			solucao = calculator.geraSolucaoAleatoria();

		return solucao;
	}

	/**
	 * Executes the Hill Climbing search with random restarts
	 */
	public int[] execute(int ciclo, boolean partidaValida)
	{
		long initTime = System.currentTimeMillis();
		this.evaluations = 0;

		int tamanhoSolucao = calculator.getSolutionSize();
		int maxEvaluations = tamanhoSolucao * tamanhoSolucao;

		int[] bestSolution = partidaValida ? geraSolucaoAleatoriaValida() : calculator.geraSolucaoAleatoria();
		double bestFitness = calculator.calculaFitness(bestSolution);
		int[] currentSolution = cloneSolution(bestSolution);

		while (evaluations < maxEvaluations)
		{
			NeighborhoodVisitorResult localSearchResult = visitNeighbors(currentSolution, maxEvaluations);

			if (localSearchResult.getStatus() == NeighborhoodVisitorStatus.FOUND_BETTER_NEIGHBOR)
			{
				currentSolution = localSearchResult.getNeighborSolution();

				if (localSearchResult.getNeighborFitness() > bestFitness)
				{
					bestSolution = cloneSolution(currentSolution);
					bestFitness = localSearchResult.getNeighborFitness();
				}
			} 
			else if (localSearchResult.getStatus() == NeighborhoodVisitorStatus.NO_BETTER_NEIGHBOR)
			{
				currentSolution = partidaValida ? geraSolucaoAleatoriaValida() : calculator.geraSolucaoAleatoria();
			}
		}

		long executionTime = System.currentTimeMillis() - initTime;
		System.out.println(calculator.imprimeSolucoes(bestSolution, ciclo, executionTime, bestFitness));
		return bestSolution;
	}
}

/**
 * Possible results of the local search phase
 */
enum NeighborhoodVisitorStatus
{
	FOUND_BETTER_NEIGHBOR, NO_BETTER_NEIGHBOR, SEARCH_EXHAUSTED
}

/**
 * Class that represents the results of the local search phase
 */
class NeighborhoodVisitorResult
{
	/**
	 * Status in the end of the local search
	 */
	private NeighborhoodVisitorStatus status;

	/**
	 * Fitness of the neighbor
	 */
	private double neighborFitness;

	/**
	 * Solution of the neighbor
	 */
	private int[] neighborSolution;

	/**
	 * Initializes a successful local search status
	 * 
	 * @param status
	 *            Status of the search, quite certainly a successful one
	 * @param fitness
	 *            Fitness of the best neighbor found
	 */
	public NeighborhoodVisitorResult(NeighborhoodVisitorStatus status, double fitness, int[] solution)
	{
		this.status = status;
		this.neighborFitness = fitness;
		neighborSolution = solution;
	}

	/**
	 * Initializes an unsuccessful local search
	 * 
	 * @param status
	 *            Status of the search: failure to find a better neighbor or
	 *            exhaustion
	 */
	public NeighborhoodVisitorResult(NeighborhoodVisitorStatus status)
	{
		this.status = status;
		this.neighborFitness = 0.0;
	}

	/**
	 * Returns the status of the local search
	 */
	public NeighborhoodVisitorStatus getStatus()
	{
		return status;
	}

	/**
	 * Return the fitness of the neighbor found, if any
	 */
	public double getNeighborFitness()
	{
		return neighborFitness;
	}

	/**
	 * Return the solution of the neighbor found, if any
	 */
	public int[] getNeighborSolution()
	{
		return neighborSolution;
	}
}
